@@ -32,6 +32,18 @@
  * @{
  */
 
+static dbus_bool_t _dbus_type_reader_greater_than              (const DBusTypeReader  *lhs,
+                                                                const DBusTypeReader  *rhs);
+
+static void       _dbus_type_writer_set_enabled           (DBusTypeWriter        *writer,
+                                                           dbus_bool_t            enabled);
+static dbus_bool_t _dbus_type_writer_write_reader_partial (DBusTypeWriter        *writer,
+                                                           DBusTypeReader        *reader,
+                                                           const DBusTypeReader  *start_after,
+                                                           int                    start_after_new_pos,
+                                                           int                    start_after_new_len,
+                                                           DBusList             **fixups);
+
 /** turn this on to get deluged in TypeReader verbose spam */
 #define RECURSIVE_MARSHAL_READ_TRACE  0
 
@@ -137,6 +149,7 @@ reader_init (DBusTypeReader    *reader,
              const DBusString  *value_str,
              int                value_pos)
 {
+  _DBUS_ZERO (*reader);
   reader->byte_order = byte_order;
   reader->finished = FALSE;
   reader->type_str = type_str;
@@ -211,7 +224,7 @@ array_reader_get_array_len (const DBusTypeReader *reader)
 
   _dbus_assert (_DBUS_ALIGN_VALUE (len_pos, 4) == (unsigned) len_pos);
   array_len = _dbus_unpack_uint32 (reader->byte_order,
-                                   _dbus_string_get_const_data_len (reader->value_str, len_pos, 4));
+                                   _dbus_string_get_const_udata_len (reader->value_str, len_pos, 4));
 
 #if RECURSIVE_MARSHAL_READ_TRACE
   _dbus_verbose ("   reader %p len_pos %d array len %u len_offset %d\n",
@@ -334,7 +347,7 @@ _dbus_type_signature_next (const char       *type_str,
   _dbus_assert (type_str != NULL);
   _dbus_assert (type_pos != NULL);
   
-  start = type_str;
+  start = (const unsigned char *)type_str;
   p = start + *type_pos;
 
   _dbus_assert (*p != DBUS_STRUCT_END_CHAR);
@@ -724,11 +737,11 @@ _dbus_type_reader_init (DBusTypeReader    *reader,
                         const DBusString  *value_str,
                         int                value_pos)
 {
-  reader->klass = &body_reader_class;
-
   reader_init (reader, byte_order, type_str, type_pos,
                value_str, value_pos);
 
+  reader->klass = &body_reader_class;
+
 #if RECURSIVE_MARSHAL_READ_TRACE
   _dbus_verbose ("  type reader %p init type_pos = %d value_pos = %d remaining sig '%s'\n",
                  reader, reader->type_pos, reader->value_pos,
@@ -749,11 +762,11 @@ _dbus_type_reader_init_types_only (DBusTypeReader    *reader,
                                    const DBusString  *type_str,
                                    int                type_pos)
 {
-  reader->klass = &body_types_only_reader_class;
-
   reader_init (reader, DBUS_COMPILER_BYTE_ORDER /* irrelevant */,
                type_str, type_pos, NULL, _DBUS_INT_MAX /* crashes if we screw up */);
 
+  reader->klass = &body_types_only_reader_class;
+
 #if RECURSIVE_MARSHAL_READ_TRACE
   _dbus_verbose ("  type reader %p init types only type_pos = %d remaining sig '%s'\n",
                  reader, reader->type_pos,
@@ -842,9 +855,9 @@ _dbus_type_reader_read_raw (const DBusTypeReader  *reader,
 {
   _dbus_assert (!reader->klass->types_only);
 
-  *value_location = _dbus_string_get_const_data_len (reader->value_str,
-                                                     reader->value_pos,
-                                                     0);
+  *value_location = _dbus_string_get_const_udata_len (reader->value_str,
+                                                      reader->value_pos,
+                                                      0);
 }
 
 /**
@@ -909,7 +922,7 @@ _dbus_type_reader_get_array_length (const DBusTypeReader  *reader)
  */
 void
 _dbus_type_reader_read_fixed_multi (const DBusTypeReader  *reader,
-                                    void                  *value,
+                                    const void           **value,
                                     int                   *n_elements)
 {
   int element_type;
@@ -943,12 +956,11 @@ _dbus_type_reader_read_fixed_multi (const DBusTypeReader  *reader,
   _dbus_assert (remaining_len <= total_len);
 
   if (remaining_len == 0)
-    *(const DBusBasicValue**) value = NULL;
+    *value = NULL;
   else
-    *(const DBusBasicValue**) value =
-      (void*) _dbus_string_get_const_data_len (reader->value_str,
-                                               reader->value_pos,
-                                               remaining_len);
+    *value = _dbus_string_get_const_data_len (reader->value_str,
+                                              reader->value_pos,
+                                              remaining_len);
 
   *n_elements = remaining_len / alignment;
   _dbus_assert ((remaining_len % alignment) == 0);
@@ -977,6 +989,7 @@ _dbus_type_reader_recurse (DBusTypeReader *reader,
                            DBusTypeReader *sub)
 {
   int t;
+  const DBusTypeReaderClass *klass = NULL;
 
   t = _dbus_first_type_in_signature (reader->type_str, reader->type_pos);
 
@@ -984,41 +997,43 @@ _dbus_type_reader_recurse (DBusTypeReader *reader,
     {
     case DBUS_TYPE_STRUCT:
       if (reader->klass->types_only)
-        sub->klass = &struct_types_only_reader_class;
+        klass = &struct_types_only_reader_class;
       else
-        sub->klass = &struct_reader_class;
+        klass = &struct_reader_class;
       break;
     case DBUS_TYPE_DICT_ENTRY:
       if (reader->klass->types_only)
-        sub->klass = &dict_entry_types_only_reader_class;
+        klass = &dict_entry_types_only_reader_class;
       else
-        sub->klass = &dict_entry_reader_class;
+        klass = &dict_entry_reader_class;
       break;
     case DBUS_TYPE_ARRAY:
       if (reader->klass->types_only)
-        sub->klass = &array_types_only_reader_class;
+        klass = &array_types_only_reader_class;
       else
-        sub->klass = &array_reader_class;
+        klass = &array_reader_class;
       break;
     case DBUS_TYPE_VARIANT:
       if (reader->klass->types_only)
         _dbus_assert_not_reached ("can't recurse into variant typecode");
       else
-        sub->klass = &variant_reader_class;
+        klass = &variant_reader_class;
       break;
     default:
       _dbus_verbose ("recursing into type %s\n", _dbus_type_to_string (t));
 #ifndef DBUS_DISABLE_CHECKS
       if (t == DBUS_TYPE_INVALID)
-        _dbus_warn_check_failed ("You can't recurse into an empty array or off the end of a message body\n");
+        _dbus_warn_check_failed ("You can't recurse into an empty array or off the end of a message body");
 #endif /* DBUS_DISABLE_CHECKS */
 
       _dbus_assert_not_reached ("don't yet handle recursing into this type");
     }
 
-  _dbus_assert (sub->klass == all_reader_classes[sub->klass->id]);
+  _dbus_assert (klass != NULL);
+  _dbus_assert (klass == all_reader_classes[klass->id]);
 
-  (* sub->klass->recurse) (sub, reader);
+  (* klass->recurse) (sub, reader);
+  sub->klass = klass;
 
 #if RECURSIVE_MARSHAL_READ_TRACE
   _dbus_verbose ("  type reader %p RECURSED type_pos = %d value_pos = %d remaining sig '%s'\n",
@@ -1428,7 +1443,7 @@ _dbus_type_reader_delete (DBusTypeReader        *reader,
   return retval;
 }
 
-/**
+/*
  * Compares two readers, which must be iterating over the same value data.
  * Returns #TRUE if the first parameter is further along than the second parameter.
  *
@@ -1436,7 +1451,7 @@ _dbus_type_reader_delete (DBusTypeReader        *reader,
  * @param rhs left-hand-side (first) parameter
  * @returns whether lhs is greater than rhs
  */
-dbus_bool_t
+static dbus_bool_t
 _dbus_type_reader_greater_than (const DBusTypeReader  *lhs,
                                 const DBusTypeReader  *rhs)
 {
@@ -1650,13 +1665,13 @@ writer_recurse_init_and_check (DBusTypeWriter *writer,
         {
           if (expected != DBUS_TYPE_INVALID)
             _dbus_warn_check_failed ("Writing an element of type %s, but the expected type here is %s\n"
-                                     "The overall signature expected here was '%s' and we are on byte %d of that signature.\n",
+                                     "The overall signature expected here was '%s' and we are on byte %d of that signature.",
                                      _dbus_type_to_string (sub->container_type),
                                      _dbus_type_to_string (expected),
                                      _dbus_string_get_const_data (writer->type_str), writer->type_pos);
           else
             _dbus_warn_check_failed ("Writing an element of type %s, but no value is expected here\n"
-                                     "The overall signature expected here was '%s' and we are on byte %d of that signature.\n",
+                                     "The overall signature expected here was '%s' and we are on byte %d of that signature.",
                                      _dbus_type_to_string (sub->container_type),
                                      _dbus_string_get_const_data (writer->type_str), writer->type_pos);
           
@@ -1716,12 +1731,12 @@ write_or_verify_typecode (DBusTypeWriter *writer,
           {
             if (expected != DBUS_TYPE_INVALID)
               _dbus_warn_check_failed ("Array or variant type requires that type %s be written, but %s was written.\n"
-                                       "The overall signature expected here was '%s' and we are on byte %d of that signature.\n",
+                                       "The overall signature expected here was '%s' and we are on byte %d of that signature.",
                                        _dbus_type_to_string (expected), _dbus_type_to_string (typecode),
                                        _dbus_string_get_const_data (writer->type_str), writer->type_pos);
             else
               _dbus_warn_check_failed ("Array or variant type wasn't expecting any more values to be written into it, but a value %s was written.\n"
-                                       "The overall signature expected here was '%s' and we are on byte %d of that signature.\n",
+                                       "The overall signature expected here was '%s' and we are on byte %d of that signature.",
                                        _dbus_type_to_string (typecode),
                                        _dbus_string_get_const_data (writer->type_str), writer->type_pos);
             _dbus_assert_not_reached ("bad type inserted somewhere inside an array or variant");
@@ -1814,7 +1829,7 @@ writer_recurse_array (DBusTypeWriter   *writer,
                                          writer->type_str,
                                          writer->u.array.element_type_pos + 1))
         {
-          _dbus_warn_check_failed ("Writing an array of '%s' but this is incompatible with the expected type of elements in the parent array\n",
+          _dbus_warn_check_failed ("Writing an array of '%s' but this is incompatible with the expected type of elements in the parent array",
                                    _dbus_string_get_const_data_len (contained_type,
                                                                     contained_type_start,
                                                                     contained_type_len));
@@ -1919,7 +1934,7 @@ writer_recurse_array (DBusTypeWriter   *writer,
           _dbus_assert (_DBUS_ALIGN_VALUE (sub->u.array.len_pos, 4) ==
                         (unsigned) sub->u.array.len_pos);
           len = _dbus_unpack_uint32 (sub->byte_order,
-                                     _dbus_string_get_const_data_len (sub->value_str,
+                                     _dbus_string_get_const_udata_len (sub->value_str,
                                                                       sub->u.array.len_pos,
                                                                       4));
 
@@ -2564,7 +2579,7 @@ writer_write_reader_helper (DBusTypeWriter       *writer,
                             (unsigned) fixup.len_pos_in_reader);
 
               old_len = _dbus_unpack_uint32 (reader->byte_order,
-                                             _dbus_string_get_const_data_len (reader->value_str,
+                                             _dbus_string_get_const_udata_len (reader->value_str,
                                                                               fixup.len_pos_in_reader, 4));
 
               if (old_len != fixup.new_len && !append_fixup (fixups, &fixup))
@@ -2627,7 +2642,7 @@ writer_write_reader_helper (DBusTypeWriter       *writer,
   return FALSE;
 }
 
-/**
+/*
  * Iterate through all values in the given reader, writing a copy of
  * each value to the writer.  The reader will be moved forward to its
  * end position.
@@ -2658,7 +2673,7 @@ writer_write_reader_helper (DBusTypeWriter       *writer,
  * @param fixups list to append #DBusArrayLenFixup if the write was partial
  * @returns #FALSE if no memory
  */
-dbus_bool_t
+static dbus_bool_t
 _dbus_type_writer_write_reader_partial (DBusTypeWriter       *writer,
                                         DBusTypeReader       *reader,
                                         const DBusTypeReader *start_after,
@@ -2719,7 +2734,7 @@ _dbus_type_writer_write_reader (DBusTypeWriter       *writer,
   return _dbus_type_writer_write_reader_partial (writer, reader, NULL, 0, 0, NULL);
 }
 
-/**
+/*
  * If disabled, a writer can still be iterated forward and recursed/unrecursed
  * but won't write any values. Types will still be written unless the
  * writer is a "values only" writer, because the writer needs access to
@@ -2728,7 +2743,7 @@ _dbus_type_writer_write_reader (DBusTypeWriter       *writer,
  * @param writer the type writer
  * @param enabled #TRUE if values should be written
  */
-void
+static void
 _dbus_type_writer_set_enabled (DBusTypeWriter   *writer,
                                dbus_bool_t       enabled)
 {
