@@ -49,13 +49,26 @@
 
 static const char *appname;
 
+static void usage (int ecode) _DBUS_GNUC_NORETURN;
+
 static void
 usage (int ecode)
 {
-  fprintf (stderr, "Usage: %s [--help] [--system | --session | --bus=ADDRESS | --peer=ADDRESS] [--dest=NAME] [--type=TYPE] [--print-reply[=literal]] [--reply-timeout=MSEC] <destination object path> <message name> [contents ...]\n", appname);
+  fprintf (stderr, "Usage: %s [--help] [--system | --session | --bus=ADDRESS | --peer=ADDRESS] [--sender=NAME] [--dest=NAME] [--type=TYPE] [--print-reply[=literal]] [--reply-timeout=MSEC] <destination object path> <message name> [contents ...]\n", appname);
   exit (ecode);
 }
 
+/* Abort on any allocation failure; there is nothing else we can do. */
+static void
+handle_oom (dbus_bool_t success)
+{
+  if (!success)
+    {
+      fprintf (stderr, "%s: Ran out of memory\n", appname);
+      exit (1);
+    }
+}
+
 static void
 append_arg (DBusMessageIter *iter, int type, const char *value)
 {
@@ -68,68 +81,68 @@ append_arg (DBusMessageIter *iter, int type, const char *value)
   double d;
   unsigned char byte;
   dbus_bool_t v_BOOLEAN;
-  
-  /* FIXME - we are ignoring OOM returns on all these functions */
+  dbus_bool_t ret;
+
   switch (type)
     {
     case DBUS_TYPE_BYTE:
       byte = strtoul (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_BYTE, &byte);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_BYTE, &byte);
       break;
 
     case DBUS_TYPE_DOUBLE:
       d = strtod (value, NULL);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_DOUBLE, &d);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_DOUBLE, &d);
       break;
 
     case DBUS_TYPE_INT16:
       int16 = strtol (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_INT16, &int16);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_INT16, &int16);
       break;
 
     case DBUS_TYPE_UINT16:
       uint16 = strtoul (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT16, &uint16);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT16, &uint16);
       break;
 
     case DBUS_TYPE_INT32:
       int32 = strtol (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_INT32, &int32);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_INT32, &int32);
       break;
 
     case DBUS_TYPE_UINT32:
       uint32 = strtoul (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT32, &uint32);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT32, &uint32);
       break;
 
     case DBUS_TYPE_INT64:
       int64 = strtoll (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_INT64, &int64);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_INT64, &int64);
       break;
 
     case DBUS_TYPE_UINT64:
       uint64 = strtoull (value, NULL, 0);
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT64, &uint64);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_UINT64, &uint64);
       break;
 
     case DBUS_TYPE_STRING:
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &value);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &value);
       break;
 
     case DBUS_TYPE_OBJECT_PATH:
-      dbus_message_iter_append_basic (iter, DBUS_TYPE_OBJECT_PATH, &value);
+      ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_OBJECT_PATH, &value);
       break;
 
     case DBUS_TYPE_BOOLEAN:
       if (strcmp (value, "true") == 0)
 	{
 	  v_BOOLEAN = TRUE;
-	  dbus_message_iter_append_basic (iter, DBUS_TYPE_BOOLEAN, &v_BOOLEAN);
+	  ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_BOOLEAN, &v_BOOLEAN);
 	}
       else if (strcmp (value, "false") == 0)
 	{
 	  v_BOOLEAN = FALSE;
-	  dbus_message_iter_append_basic (iter, DBUS_TYPE_BOOLEAN, &v_BOOLEAN);
+	  ret = dbus_message_iter_append_basic (iter, DBUS_TYPE_BOOLEAN, &v_BOOLEAN);
 	}
       else
 	{
@@ -142,6 +155,8 @@ append_arg (DBusMessageIter *iter, int type, const char *value)
       fprintf (stderr, "%s: Unsupported data type %c\n", appname, (char) type);
       exit (1);
     }
+
+  handle_oom (ret);
 }
 
 static void
@@ -150,6 +165,8 @@ append_array (DBusMessageIter *iter, int type, const char *value)
   const char *val;
   char *dupval = strdup (value);
 
+  handle_oom (dupval != NULL);
+
   val = strtok (dupval, ",");
   while (val != NULL)
     {
@@ -165,15 +182,17 @@ append_dict (DBusMessageIter *iter, int keytype, int valtype, const char *value)
   const char *val;
   char *dupval = strdup (value);
 
+  handle_oom (dupval != NULL);
+
   val = strtok (dupval, ",");
   while (val != NULL)
     {
       DBusMessageIter subiter;
       
-      dbus_message_iter_open_container (iter,
-					DBUS_TYPE_DICT_ENTRY,
-					NULL,
-					&subiter);
+      handle_oom (dbus_message_iter_open_container (iter,
+                                                    DBUS_TYPE_DICT_ENTRY,
+                                                    NULL,
+                                                    &subiter));
 
       append_arg (&subiter, keytype, val);
       val = strtok (NULL, ",");
@@ -184,7 +203,7 @@ append_dict (DBusMessageIter *iter, int keytype, int valtype, const char *value)
 	}
       append_arg (&subiter, valtype, val);
 
-      dbus_message_iter_close_container (iter, &subiter);
+      handle_oom (dbus_message_iter_close_container (iter, &subiter));
       val = strtok (NULL, ",");
     } 
   free (dupval);
@@ -242,6 +261,7 @@ main (int argc, char *argv[])
   int message_type = DBUS_MESSAGE_TYPE_SIGNAL;
   const char *type_str = NULL;
   const char *address = NULL;
+  const char *sender = NULL;
   int is_bus = FALSE;
   int session_or_system = FALSE;
 
@@ -270,14 +290,17 @@ main (int argc, char *argv[])
         }
       else if ((strstr (arg, "--bus=") == arg) || (strstr (arg, "--peer=") == arg) || (strstr (arg, "--address=") == arg))
         {
-          if (arg[2] == 'b') /* bus */
-            {
-              is_bus = TRUE;
-            }
-          else if (arg[2] == 'p') /* peer */
+          /* Check for peer first, to avoid the GCC -Wduplicated-branches
+           * warning.
+           */
+          if (arg[2] == 'p') /* peer */
             {
               is_bus = FALSE;
             }
+          else if (arg[2] == 'b') /* bus */
+            {
+              is_bus = TRUE;
+            }
           else /* address; keeping backwards compatibility */
             {
               is_bus = FALSE;
@@ -291,6 +314,16 @@ main (int argc, char *argv[])
               usage (1);
             }
         }
+      else if (strstr (arg, "--sender=") == arg)
+        {
+          sender = strchr (arg, '=') + 1;
+
+          if (sender[0] == '\0')
+            {
+              fprintf (stderr, "\"--sender=\" requires a NAME\n");
+              usage (1);
+            }
+        }
       else if (strncmp (arg, "--print-reply", 13) == 0)
 	{
 	  print_reply = TRUE;
@@ -350,6 +383,12 @@ main (int argc, char *argv[])
       usage (1);
     }
 
+  if (sender != NULL && address != NULL && !is_bus)
+    {
+      fprintf (stderr, "\"--peer\" may not be used with \"--sender\"\n");
+      exit (1);
+    }
+
   if (type_str != NULL)
     {
       message_type = dbus_message_type_from_string (type_str);
@@ -361,15 +400,30 @@ main (int argc, char *argv[])
           exit (1);
         }
     }
-  
+
   dbus_error_init (&error);
 
   if (dest && !dbus_validate_bus_name (dest, &error))
     {
       fprintf (stderr, "invalid value (%s) of \"--dest\"\n", dest);
+      dbus_error_free (&error);
       usage (1);
     }
 
+  if (sender && !dbus_validate_bus_name (sender, &error))
+    {
+      fprintf (stderr, "invalid value (%s) of \"--sender\"\n", sender);
+      dbus_error_free (&error);
+      usage (1);
+    }
+
+  if (!dbus_validate_path (path, &error))
+    {
+      fprintf (stderr, "%s\n", error.message);
+      dbus_error_free (&error);
+      exit (1);
+    }
+
   if (address != NULL)
     {
       connection = dbus_connection_open (address, &error);
@@ -399,6 +453,29 @@ main (int argc, char *argv[])
         }
     }
 
+  if (sender != NULL)
+    {
+      int ret = dbus_bus_request_name (connection, sender, DBUS_NAME_FLAG_DO_NOT_QUEUE, &error);
+      switch (ret)
+        {
+        case DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER:
+          /* success */
+          break;
+        case DBUS_REQUEST_NAME_REPLY_EXISTS:
+          fprintf (stderr, "Requested name \"%s\" already has owner\n", sender);
+          exit (1);
+        case -1:
+          fprintf (stderr, "Failed to request sender name \"%s\": %s\n", sender, error.message);
+          dbus_error_free (&error);
+          exit (1);
+        default:
+          /* This should be unreachable if the bus is compliant */
+          fprintf (stderr, "Failed to request sender name \"%s\": unexpected result code %d\n",
+                   sender, ret);
+          exit (1);
+        }
+    }
+
   if (message_type == DBUS_MESSAGE_TYPE_METHOD_CALL)
     {
       char *last_dot;
@@ -411,11 +488,28 @@ main (int argc, char *argv[])
           exit (1);
         }
       *last_dot = '\0';
-      
+
+      if (!dbus_validate_interface (name, &error))
+        {
+          /* Typically this is "Interface name was not valid: \"xxx\""
+           * so we don't need to prefix anything special */
+          fprintf (stderr, "%s\n", error.message);
+          dbus_error_free (&error);
+          exit (1);
+        }
+
+      if (!dbus_validate_member (last_dot + 1, &error))
+        {
+          fprintf (stderr, "Invalid method name: %s\n", error.message);
+          dbus_error_free (&error);
+          exit (1);
+        }
+
       message = dbus_message_new_method_call (NULL,
                                               path,
                                               name,
                                               last_dot + 1);
+      handle_oom (message != NULL);
       dbus_message_set_auto_start (message, TRUE);
     }
   else if (message_type == DBUS_MESSAGE_TYPE_SIGNAL)
@@ -430,8 +524,23 @@ main (int argc, char *argv[])
           exit (1);
         }
       *last_dot = '\0';
-      
+
+      if (!dbus_validate_interface (name, &error))
+        {
+          fprintf (stderr, "%s\n", error.message);
+          dbus_error_free (&error);
+          exit (1);
+        }
+
+      if (!dbus_validate_member (last_dot + 1, &error))
+        {
+          fprintf (stderr, "Invalid signal name: %s\n", error.message);
+          dbus_error_free (&error);
+          exit (1);
+        }
+
       message = dbus_message_new_signal (path, name, last_dot + 1);
+      handle_oom (message != NULL);
     }
   else
     {
@@ -457,13 +566,14 @@ main (int argc, char *argv[])
     {
       char *arg;
       char *c;
-      int type;
+      int type2;
       int secondary_type;
       int container_type;
       DBusMessageIter *target_iter;
       DBusMessageIter container_iter;
 
-      type = DBUS_TYPE_INVALID;
+      type2 = DBUS_TYPE_INVALID;
+      secondary_type = DBUS_TYPE_INVALID;
       arg = argv[i++];
       c = strchr (arg, ':');
 
@@ -497,9 +607,9 @@ main (int argc, char *argv[])
 	}
 
       if (arg[0] == 0)
-	type = DBUS_TYPE_STRING;
+	type2 = DBUS_TYPE_STRING;
       else
-	type = type_from_name (arg);
+	type2 = type_from_name (arg);
 
       if (container_type == DBUS_TYPE_DICT_ENTRY)
 	{
@@ -514,25 +624,25 @@ main (int argc, char *argv[])
 	  *(c++) = 0;
 	  secondary_type = type_from_name (arg);
 	  sig[0] = DBUS_DICT_ENTRY_BEGIN_CHAR;
-	  sig[1] = type;
+	  sig[1] = type2;
 	  sig[2] = secondary_type;
 	  sig[3] = DBUS_DICT_ENTRY_END_CHAR;
 	  sig[4] = '\0';
-	  dbus_message_iter_open_container (&iter,
-					    DBUS_TYPE_ARRAY,
-					    sig,
-					    &container_iter);
+          handle_oom (dbus_message_iter_open_container (&iter,
+                                                        DBUS_TYPE_ARRAY,
+                                                        sig,
+                                                        &container_iter));
 	  target_iter = &container_iter;
 	}
       else if (container_type != DBUS_TYPE_INVALID)
 	{
 	  char sig[2];
-	  sig[0] = type;
+	  sig[0] = type2;
 	  sig[1] = '\0';
-	  dbus_message_iter_open_container (&iter,
-					    container_type,
-					    sig,
-					    &container_iter);
+          handle_oom (dbus_message_iter_open_container (&iter,
+                                                        container_type,
+                                                        sig,
+                                                        &container_iter));
 	  target_iter = &container_iter;
 	}
       else
@@ -540,19 +650,20 @@ main (int argc, char *argv[])
 
       if (container_type == DBUS_TYPE_ARRAY)
 	{
-	  append_array (target_iter, type, c);
+	  append_array (target_iter, type2, c);
 	}
       else if (container_type == DBUS_TYPE_DICT_ENTRY)
 	{
-	  append_dict (target_iter, type, secondary_type, c);
+	  _dbus_assert (secondary_type != DBUS_TYPE_INVALID);
+	  append_dict (target_iter, type2, secondary_type, c);
 	}
       else
-	append_arg (target_iter, type, c);
+	append_arg (target_iter, type2, c);
 
       if (container_type != DBUS_TYPE_INVALID)
 	{
-	  dbus_message_iter_close_container (&iter,
-					     &container_iter);
+          handle_oom (dbus_message_iter_close_container (&iter,
+                                                         &container_iter));
 	}
     }
 
